#include "fireflier_client.h"
#include "sourceform.h"
#include "rules_dialog.h"
#include "fireflier_indep.h"
#include "login_dialog.h"
#include "about_dialog.h"
#include <pthread.h>
#include <qapplication.h>
#include <qpushbutton.h>
#include <qcheckbox.h>
#include <qlabel.h>
#include <qcombobox.h>
#include <qradiobutton.h>
#include <qtextedit.h>
#include <qthread.h>
#include <qlistbox.h>
#include <netdb.h>
#include <qlineedit.h>
#include <unistd.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>


char *ICMP_TYPES[]={"ECHO_REPLY", "UNKNOWN", "UNKNOWN", "DEST_UNREACH", "SOURCE_QUENCH",
"REDIRECT", "UNKNOWN", "UNKNOWN", "ECHO", "UNKNOWN", "UNKNOWN",
"TIME_EXCEEDED", "PARAMETERPROB", "TIMESTAMP", "TIMESTAMP_REPLY",
"INFO_REQUEST", "INFO_REPLY", "ADDRESS", "ADDRESS_REPLY"};

void sendSourceRequest();
void sendUSRulesRequest();
void sendIPTRulesRequest();
void sendUSdelCmd(int rulenum);
void sendIPTdelCmd(int chain, int rulenum);
int sendLogin(const char *username, const char *password);
extern int authenticated;

fireflier_client *cl;
QApplication *qApp;
rules_dialog *usr_rules;
rules_dialog *ipt_rules;
char *sourceText;
int cmdnum;

class ResolverThread : public QThread {
    char res_src[16], res_dst[16];
public:
    ResolverThread();
    void resolve(char *src, char *dest);
    void run();
};

ResolverThread::ResolverThread()
{
    res_src[0]=0;
    res_dst[0]=0;
}

void ResolverThread::resolve(char *src, char *dest)
{
    strncpy(res_src, src, 16);
    strncpy(res_dst, dest, 16);
}

void *networkThread(void *clientParameter);



void setSrcName(char *name);
void setDstName(char *name);
ResolverThread rst;

void ResolverThread::run()
{
    char src[16], dst[16];
    char buffer[4];
    struct hostent *he;
    struct in_addr addr;

    while (1)
    {
	while (res_src[0]==0)
	    sleep(1);
        strcpy(src, res_src);
        strcpy(dst, res_dst);
	cout << "Resolving: " << src << endl;;
        inet_aton(src, &addr);
	he=gethostbyaddr(&addr, sizeof(struct in_addr), AF_INET);
	if (he)
	{
	    if (!strcmp(src, res_src))
	    {
		cout << "Resolved: " << he->h_name << endl;
		setSrcName(he->h_name);
                res_src[0]=0;
	    }
	}
	else
	{
	    setSrcName("not resolvable");
	    res_src[0]=0;
	}
	cout << "Resolving: " << dst << endl;
        inet_aton(dst, &addr);
	he=gethostbyaddr(&addr, sizeof(struct in_addr), AF_INET);
	if (he)
	{
	    if (!strcmp(dst, res_dst))
	    {
		setDstName(he->h_name);
		cout << "Resolved: " << he->h_name << endl;
                res_dst[0]=0;
	    }
	}
	else
	{
	    setDstName("not resolvable");
	    res_dst[0]=0;
	}

    }
}

void setSrcName(char *name)
{
    qApp->lock();
    cl->label_ip_src->setText(name);
    qApp->unlock();
}

void setDstName(char *name)
{
    qApp->lock();
    cl->label_ip_dst->setText(name);
    cl->label_ip_dst->setText(name);
    qApp->unlock();
}

class NWThread : public QThread {
public:
    unsigned char hostip[4];
    int port;
    void run();
};

void NWThread::run()
{
    unsigned char param[6];
    memcpy(param, hostip, 4);
    param[4]=port&255;
    param[5]=port>>8;
    networkThread((void*)param);
    cout << "no connection" << endl;
    qApp->lock();
    qApp->exit(1);
    qApp->unlock();
}


/*
 *  Constructs a fireflier_client which is a child of 'parent', with the
 *  name 'name' and widget flags set to 'f'
 */
fireflier_client::fireflier_client( QWidget* parent,  const char* name, WFlags fl )
    : fireflier_main( parent, name, fl )
{
    packetAvailable=0;
}

/*
 *  Destroys the object and frees any allocated resources
 */
fireflier_client::~fireflier_client()
{
    // no need to delete child widgets, Qt does it all for us
}

/*
 * public slot
 */
void fireflier_client::fileExit()
{
    exit(0);
}

char *HEX="0123456789ABCDEF";

char *format_source(char *src, int len)
{
    char *result=new char[len*8];
    int i, pos=0;

    if (!src) return "";

    result[0]=0;

    while (pos<len)
    {
	for (int i=0;i<16;i++)
	{
	    if (i==8)
                strcat(result, "  ");
	    else if (i%4==0)
                strcat(result, " ");
	    if (i+pos<len)
	    {
		result[strlen(result)+2]=0;
		result[strlen(result)+1]=HEX[((unsigned char*)src)[i+pos]&15];
		result[strlen(result)]=HEX[((unsigned char*)src)[i+pos]>>4];
                strcat(result, " ");
	    }
	    else
                strcat(result, "   ");
	}
        strcat(result, "   ");
	for (int i=0;i<16;i++)
	{
	    if (i==8)
                strcat(result, "  ");
	    else if (i%4==0)
                strcat(result, " ");

	    if ((i+pos<len) && (src[i+pos]>=32))
	    {
                result[strlen(result)+1]=0;
		result[strlen(result)]=src[i+pos];
	    }
	    else if (i+pos<len)
                strcat(result, ".");
	}
	pos+=16;
        strcat(result, "\n");
    }
    return result;
}

void view_source_received(char *source, int len)
{
    char *formatted;
    source[len]=0;
    formatted=format_source(source, len);
    for (int i=0;formatted[i]!=0;i++)
    {
	cout << (int)formatted[i] << " ";
    }
    cout << endl;
    sourceText=formatted;
}
/*
 * public slot
 */
void fireflier_client::view_source()
{
    if (packetAvailable)
    {
	Sourceform *sf=new Sourceform(this, "View source", true);
	sendSourceRequest();
	while (!sourceText) ; // FIX THIS
        sf->source->setText(sourceText);
	sf->exec();
	delete(sourceText);
	sourceText=0;
        delete(sf);
    }
}


void view_iptables_received(char *buffer, int len)
{
    char *temp;
    char *pos, *endpos;
    char line[200];

    temp=new char[len+1];
    memcpy(temp, buffer, len);
    endpos=pos=temp;

    qApp->lock();
    ipt_rules->iptables_list->clear();
    qApp->unlock();

    while (endpos && (endpos<temp+strlen(temp))) {
	endpos=(char*)memchr(pos, '\n', strlen(temp)-(pos-temp));
	if (endpos && (endpos<temp+strlen(temp)))
	{
	    memcpy(line, pos, endpos-pos);
	    line[endpos-pos]=0;
            qApp->lock();
	    ipt_rules->iptables_list->insertItem(line);
	    qApp->unlock();
	    cout << "::" << line << endl;
	}
	pos=endpos+1;
    }
    delete(temp);
}

void fireflier_client::view_iptables()
{
    if (!ipt_rules)
    {
	ipt_rules=new rules_dialog();
	connect(ipt_rules->delete_rule , SIGNAL( clicked() ), this, SLOT( del_iptrule_clicked() ) );
	ipt_rules->show();
    }
    else if (!ipt_rules->isVisible())
    {
        ipt_rules->iptables_list->clear();
	ipt_rules->show();
    }
    sendIPTRulesRequest();
}

void view_userspace_received(char *buffer, int len)
{
    char *temp;
    char *pos, *endpos;
    char line[200];

    temp=new char[len+1];
    memcpy(temp, buffer, len);
    endpos=pos=temp;

    qApp->lock();
    usr_rules->iptables_list->clear();
    qApp->unlock();

    while (endpos && (endpos<temp+strlen(temp))) {
	endpos=(char*)memchr(pos, '\n', strlen(temp)-(pos-temp));
	if (endpos && (endpos<temp+strlen(temp)))
	{
	    memcpy(line, pos, endpos-pos);
	    line[endpos-pos]=0;
            qApp->lock();
	    usr_rules->iptables_list->insertItem(line);
	    qApp->unlock();
	    cout << "::" << line << endl;
	}
	pos=endpos+1;
    }
    delete(temp);
}

void fireflier_client::view_userspace()
{
    if (!usr_rules)
    {
	usr_rules=new rules_dialog();
	connect(usr_rules->delete_rule , SIGNAL( clicked() ), this, SLOT( del_usrule_clicked() ) );
	usr_rules->show();
    }
    else if (!usr_rules->isVisible())
    {
        usr_rules->iptables_list->clear();
	usr_rules->show();
    }
    sendUSRulesRequest();
}

void fireflier_client::helpAbout()
{
    about *ad=new about();
    ad->show();
}

void clearInterface()
{
    QString temp="not available";
    cout << "Clearing Interface..." << endl;

    cl->cb_ip_src->setText(temp); cl->cb_ip_src->setEnabled(false); cl->cb_ip_src->setChecked(false);
    cl->cb_ip_dst->setText(temp); cl->cb_ip_dst->setEnabled(false); cl->cb_ip_dst->setChecked(false);
    cl->cb_port_src->setText(temp); cl->cb_port_src->setEnabled(false); cl->cb_port_src->setChecked(false);
    cl->cb_port_dst->setText(temp); cl->cb_port_dst->setEnabled(false); cl->cb_port_dst->setChecked(false);
    cl->cb_interface_in->setText(temp); cl->cb_interface_in->setEnabled(false); cl->cb_interface_in->setChecked(false);
    cl->cb_interface_out->setText(temp); cl->cb_interface_out->setEnabled(false); cl->cb_interface_out->setChecked(false);
    cl->cb_protocol->setText(temp); cl->cb_protocol->setEnabled(false);cl->cb_protocol->setChecked(false);
    cl->cb_conntrack->setEnabled(false); cl->cb_conntrack->setChecked(false);
    cl->cb_program->setText(temp); cl->cb_program->setEnabled(false); cl->cb_program->setChecked(false);

    cl->cb_this_packet->setEnabled(false); cl->cb_this_packet->setChecked(true);
    cl->cb_queued_packets->setEnabled(false); cl->cb_queued_packets->setChecked(false);
    cl->cb_create_rule->setEnabled(false); cl->cb_create_rule->setChecked(false);

    cl->ch_timeout->setEnabled(false);

    cl->label_mac_addr->setText(temp); cl->label_mac_addr->setEnabled(false);
    cl->label_len->setText(temp); cl->label_len->setEnabled(false);
    cl->label_tcp_flags->setText(temp); cl->label_tcp_flags->setEnabled(false);
    cl->label_icmp_type->setText(temp); cl->label_icmp_type->setEnabled(false);
    cl->label_arrived->setText(temp); cl->label_arrived->setEnabled(false);

    cl->label_ip_src->setText(temp); cl->label_ip_src->setEnabled(false);
    cl->label_ip_dst->setText(temp); cl->label_ip_dst->setEnabled(false);
    cl->label_chain->setText(temp); cl->label_chain->setEnabled(false);
    cout << "-Clearing Interface..." << endl;
}

void activateInterface()
{
    QString temp;
    cout << "activating interface..." << endl;
    cl->cb_ip_src->setEnabled(true);
    cl->cb_ip_dst->setEnabled(true);
    cl->cb_protocol->setEnabled(true);
    cl->cb_conntrack->setEnabled(true);

    cl->cb_this_packet->setEnabled(true);
    cl->cb_queued_packets->setEnabled(true);
    cl->cb_create_rule->setEnabled(true);

    cl->ch_timeout->setEnabled(true);

    cl->label_len->setEnabled(true);
    cl->label_arrived->setEnabled(true);

    temp="resolving...";
    cl->label_ip_src->setText(temp); cl->label_ip_src->setEnabled(true);
    cl->label_ip_dst->setText(temp); cl->label_ip_dst->setEnabled(true);
    cl->label_chain->setEnabled(true);
}

void fireflier_client::accept_clicked()
{
    if (packetAvailable)
    {
        cout << "accepting" << endl;
	int id=0;
	int timeout=0;
        int ipsrc, ipdst, portsrc, portdst, intin, intout, prot, conn, prog;
	if (cb_this_packet->isChecked()) id=1;
	if (cb_queued_packets->isChecked()) id=2;
	if (cb_create_rule->isChecked()) id=3;

	if (ch_timeout->currentItem()==1)
	    timeout=5*60;
	else if (ch_timeout->currentItem()==2)
            timeout=60*60;
	else if (ch_timeout->currentItem()==3)
            timeout=24*60*60;

	ipsrc=cb_ip_src->isChecked();
	ipdst=cb_ip_dst->isChecked();
	portsrc=cb_port_src->isChecked();
	portdst=cb_port_dst->isChecked();
	intin=cb_interface_in->isChecked();
	intout=cb_interface_out->isChecked();
	prot=cb_protocol->isChecked();
	conn=cb_conntrack->isChecked();
        prog=cb_program->isChecked();

        clearInterface();
	packetAvailable=0;
	sendCmd(id, ipsrc, ipdst, portsrc, portdst, intin, intout,
		0, // reserved for mac address filtering
		prot, conn, prog, timeout,
		1 // accept
	       );
	if (ipt_rules && ipt_rules->isVisible())
            sendIPTRulesRequest();
	if (usr_rules && usr_rules->isVisible())
            sendUSRulesRequest();
    }
}

void fireflier_client::deny_clicked()
{
    if (packetAvailable)
    {
        cout << "denying" << endl;
	int id=0;
	int timeout=0;
        int ipsrc, ipdst, portsrc, portdst, intin, intout, prot, conn, prog;
	if (cb_this_packet->isChecked()) id=1;
	if (cb_queued_packets->isChecked()) id=2;
	if (cb_create_rule->isChecked()) id=3;

	if (ch_timeout->currentItem()==1)
	    timeout=5*60;
	else if (ch_timeout->currentItem()==2)
            timeout=60*60;
	else if (ch_timeout->currentItem()==3)
            timeout=24*60*60;

	ipsrc=cb_ip_src->isChecked();
	ipdst=cb_ip_dst->isChecked();
	portsrc=cb_port_src->isChecked();
	portdst=cb_port_dst->isChecked();
	intin=cb_interface_in->isChecked();
	intout=cb_interface_out->isChecked();
	prot=cb_protocol->isChecked();
	conn=cb_conntrack->isChecked();
        prog=cb_program->isChecked();

        clearInterface();
	packetAvailable=0;
	sendCmd(id, ipsrc, ipdst, portsrc, portdst, intin, intout,
		0, // reserved for mac address filtering
		prot, conn, prog, timeout,
		0 // accept
	       );
	if (ipt_rules && ipt_rules->isVisible())
            sendIPTRulesRequest();
	if (usr_rules && usr_rules->isVisible())
            sendUSRulesRequest();
    }
}

void setLabels(packet *p)
{
    char buffer[50];
    char buffer2[50];
    cout << "Setting labels..." << endl;

    activateInterface();
    char temp[50];
    if (p->ip_src)
	cl->packetAvailable=1;
    else
	cl->packetAvailable=0;

    sprintf(buffer, "%i.%i.%i.%i",
            ((unsigned char*)&p->ip_src)[0],
	    ((unsigned char*)(&p->ip_src))[1],
	    ((unsigned char*)(&p->ip_src))[2],
	    ((unsigned char*)(&p->ip_src))[3]);
    strcpy(temp, "Source IP=");
    strcat(temp, buffer);
    cl->cb_ip_src->setText(temp);
    sprintf(buffer2, "%i.%i.%i.%i",
	    ((unsigned char*)(&p->ip_dst))[0],
	    ((unsigned char*)(&p->ip_dst))[1],
	    ((unsigned char*)(&p->ip_dst))[2],
	    ((unsigned char*)(&p->ip_dst))[3]);
    strcpy(temp, "Destination IP=");
    strcat(temp, buffer2);
    rst.resolve(buffer, buffer2);
    cl->cb_ip_dst->setText(temp);

    if (p->port_src)
    {
	cl->cb_port_src->setEnabled(true);
	sprintf(buffer, "Source Port=%i", p->port_src);
	strcpy(temp, buffer);
	cl->cb_port_src->setText(temp);
    }

    if (p->port_dst)
    {
	cl->cb_port_dst->setEnabled(true);
	sprintf(buffer, "Destination Port=%i", p->port_dst);
	strcpy(temp, buffer);
	cl->cb_port_dst->setText(temp);
    }

    strcpy(buffer, "");
    switch (p->protocol)
    {
    case 1:
	strcpy(buffer, "Protocol=ICMP");
        break;
    case 6:
	strcpy(buffer, "Protocol=TCP");
        break;
    case 17:
	strcpy(buffer, "Protocol=UDP");
        break;
    case 47:
	strcpy(buffer, "Protocol=GRE");
	break;
    case 0:
	sprintf(buffer, "Protocol=UNKNOWN(%d)", p->protocol);
	break;

    }
    strcpy(temp, buffer);
    cl->cb_protocol->setText(temp);

    if (p->programname && strcmp((char*)p->programname, ""))
    {
	strcpy(temp, p->programname);
	cl->cb_program->setEnabled(true);
	cl->cb_program->setText(temp);
    }

    if (strcmp((char*)p->interface_in, ""))
    {
	cl->cb_interface_in->setEnabled(true);
	strcpy(temp, (char*)p->interface_in);
	strcpy(buffer, "Incoming interface=");
        strcat(buffer, temp);
	cl->cb_interface_in->setText(buffer);
    }

    if (strcmp((char*)p->interface_out, ""))
    {
	cl->cb_interface_out->setEnabled(true);
	strcpy(temp, (char*)p->interface_out);
	strcpy(buffer, "Outgoing interface=");
        strcat(buffer, temp);
	cl->cb_interface_out->setText(buffer);
    }

    switch (p->hook)
    {
    case 1:
	strcpy(temp,"Chain=INPUT");
	break;
    case 2:
	strcpy(temp,"Chain=FORWARD");
	break;
    case 3:
	strcpy(temp,"Chain=OUTPUT");
	break;
    default:
	strcpy(temp,"not available");
        break;
    }
    cl->label_chain->setText(temp);

    sprintf(buffer, "len=%d", p->len);
    strcpy(temp, buffer);
    cl->label_len->setText(temp);
    cl->label_len->setEnabled(temp);

    if (p->mac_addrlen>0)
    {
	sprintf(buffer, "mac=%2x:%2x:%2x:%2x:%2x:%2x", p->mac_addr[0],p->mac_addr[1],
		p->mac_addr[2],p->mac_addr[3],
		p->mac_addr[4],p->mac_addr[5]);
	strcpy(temp, buffer);
	cl->label_mac_addr->setText(temp);
	cl->label_mac_addr->setEnabled(true);
    }
    strcpy(buffer, "tcp_flags=");
    if (p->tcp_flags&1)
        strcat(buffer, "FIN ");
    if (p->tcp_flags&2)
        strcat(buffer, "SYN ");
    if (p->tcp_flags&4)
        strcat(buffer, "ACK ");
    if (p->tcp_flags&8)
	strcat(buffer, "RST ");
    if (strcmp(buffer, "tcp_flags=") && (p->protocol==6))
    {
	strcpy(temp, buffer);
	cl->label_tcp_flags->setText(temp);
	cl->label_tcp_flags->setEnabled(true);
    }
    if (p->protocol==1)
    {
	strcpy(temp, "ICMP_TYPE=");
	strcat(temp, ICMP_TYPES[p->icmp_type]);
	cl->label_icmp_type->setText(temp);
	cl->label_icmp_type->setEnabled(true);
    }

    strcpy(buffer, "Packet arrived at ");
    strcat(buffer, ctime(&((long)p->arrived)));
    buffer[strlen(buffer)-1]=0;

    cl->label_arrived->setText(buffer);

    cout << "-Setting labels..." << endl;
}

void fireflier_client::del_usrule_clicked()
{
    int rulenum;

    rulenum=usr_rules->iptables_list->currentItem();
    if (rulenum>=0)
    {
        sendUSdelCmd(rulenum);
    }
}

void fireflier_client::del_iptrule_clicked()
{
    int rulenum;
    int chain=-1;

    rulenum=ipt_rules->iptables_list->currentItem();
    if (rulenum>=0)
    {
	if (strncmp(ipt_rules->iptables_list->currentText(), "Chain ", 6))
	{
	    int i=0;
	    while (i<rulenum)
	    {
		i++;
		if (!strncmp("Chain INPUT", ipt_rules->iptables_list->item(rulenum-i)->text(),11))
		{
		    chain=1;
                    break;
		}
		if (!strncmp("Chain OUTPUT", ipt_rules->iptables_list->item(rulenum-i)->text(),12))
		{
		    chain=3;
                    break;
		}
		if (!strncmp("Chain FORWARD", ipt_rules->iptables_list->item(rulenum-i)->text(),13))
		{
		    chain=2;
                    break;
		}
	    }
            rulenum=i-1;
	}
	if (chain>=1)
	    sendIPTdelCmd(chain, rulenum);
    }
}


void usage()
{
    cout << "Usage: fireflier-client-qt [SERVER_IP [SERVER_PORT]]" << endl;
    cout << "  Default: fireflier-client-qt 127.0.0.1 1133" << endl;
}

int main(int argc, char **argv)
{
    qApp=new QApplication( argc, argv );
    NWThread qt;

    if (argc>1)
    {
	if (sscanf(argv[1], "%d.%d.%d.%d", qt.hostip, qt.hostip+1, qt.hostip+2, qt.hostip+3)!=4)
	{
	    usage();
            exit(1);
	}
    }
    else
    {
        qt.hostip[0]=127;
        qt.hostip[1]=0;
        qt.hostip[2]=0;
        qt.hostip[3]=1;
    }

    if (argc>2)
    {
	qt.port=atoi(argv[2]);
	if ((qt.port<=0) || (qt.port>65535))
	{
	    usage();
            exit(1);
	}
    }
    else
        qt.port=1133;

    sourceText=0;
    ipt_rules=0;
    usr_rules=0;
    cl=new fireflier_client();

    qApp->setMainWidget(cl);
    rst.start();

//    pthread_t nwThread;
    qt.start();

    login_dialog ld;
    authenticated=0;
    do {
	ld.exec();
	if (ld.result()==QDialog::Rejected)
	    exit(1);
	sendLogin(ld.input_username->text(), ld.input_password->text());
	while (authenticated==-1)
	    sleep(1);
    } while (authenticated!=1);
//    pthread_create(&nwThread, NULL, networkThread, (void*)&sp);

    cl->show();
    qApp->exec();
    return 0;
}
