#ifndef backend_h
#define backend_h

#include <linux/netfilter.h>
#include <netinet/in.h>
#include <syslog.h>
#include <openssl/ssl.h>
#include <net/if.h>
extern "C" {
#include <libipq/libipq.h>
}
#include <stdio.h>
#include <iostream.h>
#include <time.h>
#include <signal.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <linux/udp.h>
#include <linux/icmp.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <pthread.h>
#include "queue.h" // a simple queue which supports iterator, maximum queue length
#include "rules.h" // manage iptables rules
#include "ports.h" // resolve localip/localport to programname
#include "pam.h" // authentication

// maximum number of packets remaining in queue.
// if this queue is full, further packets are dropped
#define QUEUE_LENGTH 50
// libipq buffer size
#define BUFSIZE 16384

class packet {
public:
    unsigned long packet_id; // unique packet id, needed for accepting or dropping a packet (libipq)
    unsigned long arrived; // time the packet arrived (in seconds)
    unsigned char interface_in[IFNAMSIZ]; // name of incoming interface
    unsigned char interface_out[IFNAMSIZ]; // name of outgoing interface
    unsigned short int len; // length of packet data
    unsigned long ip_src; // source ip
    unsigned long ip_dst; // destination ip
    unsigned int protocol:8; // protocol id (see rules.h)

    unsigned short int port_src; // source port
    unsigned short int port_dst; // destination port

    unsigned short int tcp_flags; // tcp flags: ACK, FIN, RST, SYN, ...

    unsigned int icmp_type:8; // type of icmp packet

    unsigned int mac_addrlen:8; // real length of mac address 
    unsigned char mac_addr[8]; // mac is maximum 8 bytes
    unsigned int hook; // chain (INPUT=1, FORWARD=2, OUTPUT=3)
    char *data; // packet data
    char *programname; // program using this packets localport/localip
    ~packet();
};

class tentry {
public:
    char *buf; // which elements are considered important by rule (1= important, 0=not important)
    unsigned long timeout; // timeout if a rule entry
    int action; // action to be taken
    packet *pack; // copy of corresponding packet
    tentry(char *buf, packet *ptr, int act); 
    ~tentry();
};
// queue containing rules which have timeout
class queue *timeoutqueue;
// queue of userspace rules
class queue *userspacequeue;

// Commands in network communication (=first byte of each message)
// get source of packet
#define CMD_SOURCE 1
// get iptables rules
#define CMD_RULES 2
// allow packet(s)/create rule allowing packets
#define CMD_ALLOW 3
// deny packet(s)/create rule denying packets
#define CMD_DENY 4
// delete iptables rule
#define CMD_DELETE_RULE 5
// get userpace rules
#define CMD_URULES 6
// delete userspace rule
#define CMD_DELETE_URULE 7
// authenticate
#define CMD_AUTH 8

// queue of "pending packets"
class queue *packetqueue;

// currently active packet
packet *currentPacket;

// is there a client connected
int clientConnected=0;
// socket of the client
int clientSocket=-1;
// is the connected client authenticated? otherwise no commands except auth will be accepted
int clientAuthenticated=0; 

struct ipq_handle *h;

// mutual exclusion for queues
pthread_mutex_t lock_mutex;

SSL *ssl;
SSL_CTX *ssl_ctx;


extern int errno;
// handler for SIGINT, SIGTERM
void finalize(int sig);
void saveuserspacerules();

// default values for configs
char *Config_ssl_file="/etc/fireflier/fireflier.pem";
char *Config_save_file="/var/lib/fireflier/usrules.dat";
int Config_port=1133;
int Config_create_queue_rules=1;
char *Config_ssl_password="password";
char *Config_client_ip="127.0.0.1";

#endif